#  doc.py
#
#  Simple function to fetch a __doc__ attribute and do formatting based
#  on the 'recommended' style of writing docstrings (Python Tutorial).
#

"""Provides formatting for documentation strings on Python objects.

This module provides three functions and two variables to assist with
handling docstrings.  The two variables are:

    INPUT_TABWIDTH	- Width that the TAB character represents on
			  input.  The initial width is 8 on all systems.

    OUTPUT_TABWIDTH	- Width that the TAB character represents on
			  output.  The initial width is 8 except under
			  MacOS, where it is 4.

    NO_DOCSTRING_FOUND	- String to print if there is no docstring attribute
			  defined or if it is None.

The three functions provide formatting, extraction & printing, and a
convenience function for removing an indentation from a string.  They are:

    formatDocString()   - Given a documentation string, format it by
			  removing indentation according to guidelines
			  stated in the Python Tutorial if it looks as
			  if the string were written with the guidelines
			  in mind.  If not, no formatting is performed.
			  The result is returned to the caller.  No other
			  transformations are made on the string (i.e.,
			  paragraphs are not re-wrapped).

    doc()		- The high-level interface for interactive use.
			  This function takes the object to print
			  documentation for as an argument.  If a __doc__
			  attribute can be extracted and is not None,
			  the string representation is formatted and
			  displayed using "print."  If this fails, the
			  value of NO_DOCSTRING_FOUND is displayed.

    stripIndentation()  - Function to remove at most a given amount of
			  indentation from a string, specified in columns.
			  A string and integer are passed as parameters.
			  The integer specifies the width of indentation
			  to be removed.  If the string has less
			  indentation than the integer specifies, only the
			  leading whitespace is removed.  A TAB in the
			  string's indentation may be replaced by spaces
			  if only part of the indentation created by the
			  TAB should be removed.  Portions of the string
			  not required to be changed are not affected:
			  TAB characters are not globally converted to
			  spaces.  Only the first line is affected.
"""


INPUT_TABWIDTH	= 8
import os
OUTPUT_TABWIDTH	= ((os.name == 'mac') and 4) or 8
del os

DOCSTRING_SEPARATOR = '\n\n'
NO_DOCSTRING_FOUND  = 'Documentation string not present.'



def formatDocString(initstring):
    """Format a documentation string.

    formatDocString(string) ==> string

    Given a documentation string, format it by removing indentation
    according to guidelines stated in the Python Tutorial if it looks
    as if the string were written with the guidelines in mind.  If
    not, no formatting is performed.  The result is returned to the
    caller.  No other transformations are made on the string (i.e.,
    paragraphs are not re-wrapped).
    """
    from string import index, strip, whitespace, joinfields, splitfields

    initstring = strip(initstring)
    #  This 'try' protects against failure in the 'index()' below.
    try:
	#  Pull off the first line:
	output    = initstring[:index(initstring, '\n')] + '\n'
	docstring = initstring[len(output):]

	#  If the next line is blank:
	if docstring[0] == '\n':
	    output = output + '\n'
	    pos = 0
	    #  Skip extra newlines
	    while docstring[pos] == '\n':
		pos = pos + 1
	    docstring = docstring[pos:]

	    indent = 0
	    while docstring and docstring[0] in whitespace:
		ch = docstring[0]
		docstring = docstring[1:]

		if ch == '\n':
		    indent = 0
		elif ch == '\t':
		    indent = indent + INPUT_TABWIDTH
		    indent = indent - (indent % INPUT_TABWIDTH)
		else:
		    indent = indent + 1

	    if (not docstring) or (not indent):
		return output + docstring

	    #  We have an indentation offset, use it:
	    text = splitfields(docstring, '\n')
	    for i in range(0, len(text)):
		text[i] = stripIndentation(text[i], indent)

	    return output + joinfields(text, '\n')
	else:
	    return initstring
    except ValueError:
	return initstring



def sdoc(obj):
    """Retrieve, format and print an abbreviated documentation string.

    sdoc(object) ==> None
    """
    try:
	docstring = obj.__doc__

	if docstring is None:
	    print NO_DOCSTRING_FOUND
	    return

	docstring = formatDocString(str(docstring))
	import string
	docstring = string.splitfields(docstring, '\n')
	if len(docstring) >= 2:
	    if docstring[1] == '':
		print docstring[0]
	    else:
		#  second line not blank
		print docstring[0], '...'
	else:
	    print docstring[0]

    except AttributeError:
	print NO_DOCSTRING_FOUND



def doc(obj):
    """Retrieve, format and print a documentation string.

    doc(object) ==> None

    This function extracts a __doc__ attribute from an object and removes
    extraneous indentation introduced in emacs python-mode to make the
    string "print worthy."  This does not modify the actual __doc__
    attribute of the object passed in, nor does it re-wrap the text of the
    input lines to a particular width.  Documentation strings that appear
    to follow Guido's recommendations in the Python Tutorial will have
    their indentation "normalized" while all other docstrings will be
    unaffected.  The result is printed on the standard output stream using
    a "print" statement.
    """
    try:
	docstring = obj.__doc__
    except AttributeError:
	print NO_DOCSTRING_FOUND
    else:
	if docstring is None:
	    print NO_DOCSTRING_FOUND
	    return
	print formatDocString(str(docstring))



def stripIndentation(str, indent):
    """Remove at most a given amount of indentation from a string.

    stripIndentation(str, indent) ==> string

    Function to remove at most a given amount of indentation from a
    string, specified in columns.  A string and integer are passed as
    parameters.  The integer specifies the width of indentation to be
    removed.  If the string has less indentation than the integer
    specifies, only the leading whitespace is removed.  A TAB in the
    string's indentation may be replaced by spaces if only part of the
    indentation created by the TAB should be removed.  Portions of the
    string not required to be changed are not affected: TAB characters
    are not globally converted to spaces.  Only the first line is
    affected.
    """
    from string import whitespace

    pos = removed = 0
    length = len(str)

    while removed < indent and pos < length:
	ch = str[pos]
	if ch == '\t':
	    width = INPUT_TABWIDTH - (removed % INPUT_TABWIDTH)
	    if removed + width > indent:
		str = (' ' * (removed + width - indent)) + str[pos+1:]
		pos = 0
		break
	    else:
		pos = pos + 1
	    removed = removed + width
	elif ch == '\n':
	    break
	elif ch in whitespace:
	    removed = removed + 1
	    pos     = pos + 1

    return str[pos:]


#  This data element is used in in the functions below.
#
import types
_rdoc_types = (types.ClassType, types.MethodType, types.InstanceType)
del types


def rdoc(obj):
    """Prints the documentation of a class or method recursively.

    rdoc(object) ==> None

    Object passed to rdoc() which are not classes or methods are simply
    passed on to doc().
    """
    t = type(obj)
    if t not in _rdoc_types:
	doc(obj)
	return
    docEntries = fetchRecursiveDocs(obj)
    print formatRecursiveDocs(docEntries)


def fetchRecursiveDocs(obj):
    """Return all docstrings associated with a class or method.
    """
    import types
    t = type(obj)
    if t is types.MethodType:
	sel = obj.im_self
	if sel is not None:
	    cls = sel.__class__
	else:
	    cls = obj.im_class
	nam = obj.im_func.__name__
	docs = fetchMethodDocs(cls, nam)
    elif t is types.ClassType:
	docs = fetchClassDocs(obj)
    elif t is types.InstanceType:
	docs = fetchClassDocs(obj.__class__)
    return docs


def fetchClassDocs(classobj):
    """Return the documentation tree for a class object.
    """
    return [classobj.__doc__] + map(fetchClassDocs, classobj.__bases__)


def firstClassDoc(cls):
    """Locates and return the first docstring defined for a class object.
    """
    d = cls.__doc__
    for c in cls.__bases__:
	d = firstClassDoc(c)
	if d:
	    break
    return d


def fetchMethodDocs(cls, meth):
    """Locates all docstrings defined for a class method.
    """
    if cls.__dict__.has_key(meth):
	d = cls.__dict__[meth].__doc__
    else:
	d = None

    names = [meth] * len(cls.__bases__)
    return [d] + map(fetchMethodDocs, cls.__bases__, names)


def firstMethodDoc(cls, meth):
    """Locates and returns the first docstring defined for a class method.
    """
    if cls.__dict__.has_key(meth):
	d = cls.__dict__[meth].__doc__
    else:
	d = None

    if not d:
	for c in cls.__bases__:
	    d = firstMethodDoc(c, meth)
	    if d:
		break
    return d


def firstRecursiveDoc(docs):
    """Locates the first non-empty docstring in a documentation tree.
    """
    result = ''
    if docs:
	if docs[0]:
	    result = docs[0]
	else:
	    for d in docs[1]:
		result = firstRecursiveDoc(d)
		if result:
		    break
    return result


def formatRecursiveDocs(entries):
    """Creates a formatted representation of a documentation tree.
    """
    lst = []
    for e in entries:
	if type(e) is type(''):
	    lst.append(formatDocString(e))
	elif type(e) is type([]):
	    l = formatRecursiveDocs(e)
	    if l:
		lst.append(l)
    #  Join the entries together:
    from string import joinfields
    return joinfields(lst, DOCSTRING_SEPARATOR)


#
#  end of doc.py
